@@ -1,7 +1,7 @@
 /* -*- c-file-style: "ruby"; indent-tabs-mode: nil -*- */
 /*
- *  Copyright (C) 2011  Ruby-GNOME2 Project Team
- *  Copyright (C) 2002-2004  Ruby-GNOME2 Project Team
+ *  Copyright (C) 2011-2019  Ruby-GNOME Project Team
+ *  Copyright (C) 2002-2004  Ruby-GNOME Project Team
  *  Copyright (C) 2002,2003  Masahiro Sakai
  *
  *  This library is free software; you can redistribute it and/or
@@ -31,6 +31,39 @@ VALUE rbgobj_signal_wrap(guint sig_id);
 
 /**********************************************************************/
 
+static const rb_data_type_t rg_glib_signal_type = {
+    "GLib::Signal",
+    {
+        NULL,
+        NULL,
+    },
+    NULL,
+    NULL,
+    RUBY_TYPED_FREE_IMMEDIATELY,
+};
+
+VALUE
+rbgobj_signal_new(guint id)
+{
+    VALUE rb_query;
+    GSignalQuery *query;
+
+    rb_query = TypedData_Make_Struct(RG_TARGET_NAMESPACE,
+                                     GSignalQuery,
+                                     &rg_glib_signal_type,
+                                     query);
+    g_signal_query(id, query);
+    return rb_query;
+}
+
+static GSignalQuery *
+rbgobj_signal_get_raw(VALUE rb_signal)
+{
+    GSignalQuery *query;
+    TypedData_Get_Struct(rb_signal, GSignalQuery, &rg_glib_signal_type, query);
+    return query;
+}
+
 static VALUE signal_func_table;
 
 void
@@ -51,6 +84,28 @@ rbgobj_get_signal_func(guint signal_id)
     return func;
 }
 
+static VALUE signal_call_func_table;
+
+void
+rbgobj_set_signal_call_func(VALUE klass,
+                            const gchar *signal_name,
+                            RGClosureCallFunc func)
+{
+    VALUE obj = Data_Wrap_Struct(rb_cData, NULL, NULL, func);
+    guint signal_id = g_signal_lookup(signal_name, CLASS2GTYPE(klass));
+    rb_hash_aset(signal_call_func_table, UINT2NUM(signal_id), obj);
+}
+
+RGClosureCallFunc
+rbgobj_get_signal_call_func(guint signal_id)
+{
+    RGClosureCallFunc func = NULL;
+    VALUE func_obj = rb_hash_aref(signal_call_func_table, UINT2NUM(signal_id));
+    if (!NIL_P(func_obj))
+        Data_Get_Struct(func_obj, void, func);
+    return func;
+}
+
 /**********************************************************************/
 
 static VALUE eNoSignalError;
@@ -75,7 +130,7 @@ accumulator_func(G_GNUC_UNUSED GSignalInvocationHint *ihint,
         continue_emission = RVAL2CBOOL(rb_ary_entry(tmp, 0));
         val = rb_ary_entry(tmp, 1);
     } else {
-        val = tmp;        
+        val = tmp;
     }
     rbgobj_rvalue_to_gvalue(val, return_accu);
 
@@ -101,11 +156,11 @@ rbg_rval2gtypes_body(VALUE value)
 }
 
 static G_GNUC_NORETURN VALUE
-rbg_rval2gtypes_rescue(VALUE value)
+rbg_rval2gtypes_rescue(VALUE value, VALUE error)
 {
     g_free(((struct rval2gtypes_args *)value)->result);
 
-    rb_exc_raise(rb_errinfo());
+    rb_exc_raise(error);
 }
 
 static GType *
@@ -142,7 +197,7 @@ rbg_rval2gtypes_accept_nil(volatile VALUE *value, long *n)
 #define RVAL2GTYPES_ACCEPT_NIL(value, n) rbg_rval2gtypes_accept_nil(&(value), &(n))
 
 static VALUE
-gobj_s_signal_new(int argc, VALUE* argv, VALUE self)
+gobj_s_define_signal(int argc, VALUE* argv, VALUE self)
 {
     const RGObjClassInfo *cinfo = rbgobj_lookup_class(self);
     VALUE rbsignal_name, rbsignal_flags, accumulator, rbreturn_type, params;
@@ -153,6 +208,7 @@ gobj_s_signal_new(int argc, VALUE* argv, VALUE self)
     GType *param_types;
     long n_params;
     guint signal;
+    VALUE rb_signal;
 
     rb_scan_args(argc, argv, "4*",
                  &rbsignal_name, &rbsignal_flags, &accumulator, &rbreturn_type, &params);
@@ -165,20 +221,23 @@ gobj_s_signal_new(int argc, VALUE* argv, VALUE self)
         rbsignal_name = rb_str_new2(rb_id2name(SYM2ID(rbsignal_name)));
     signal_name = RVAL2CSTR(rbsignal_name);
 
-    signal_flags = NUM2INT(rbsignal_flags);
+    signal_flags = RVAL2GFLAGS(rbsignal_flags, G_TYPE_SIGNAL_FLAGS);
 
     {
+        VALUE rb_method_name;
         VALUE proc;
         ID method_id;
 
-        method_id = rb_to_id(rb_str_concat(rb_str_new2(default_handler_method_prefix), rbsignal_name));
+        rb_method_name =
+            rb_str_concat(rb_str_new_cstr(default_handler_method_prefix),
+                          rbsignal_name);
+        method_id = rb_to_id(rb_method_name);
 
         proc = rb_funcall(mMetaInterface, rb_intern("signal_callback"), 2,
                           self, ID2SYM(method_id));
 
         class_closure = g_rclosure_new(proc, Qnil, NULL);
-        /* TODO: Should this be done even if something below it fails? */
-        g_rclosure_attach(class_closure, self);
+        g_rclosure_set_tag(class_closure, RVAL2CSTR(rb_method_name));
     }
 
     return_type = rbgobj_gtype_get(rbreturn_type);
@@ -192,7 +251,7 @@ gobj_s_signal_new(int argc, VALUE* argv, VALUE self)
                            NIL_P(accumulator) ? NULL : (gpointer)accumulator,
                            NULL, /* c_marshaller */
                            return_type,
-                           n_params,
+                           (guint)n_params,
                            param_types);
 
     g_free(param_types);
@@ -200,10 +259,13 @@ gobj_s_signal_new(int argc, VALUE* argv, VALUE self)
     if (!signal)
         rb_raise(rb_eRuntimeError, "g_signal_newv failed");
 
+    rb_signal = rbgobj_signal_new(signal);
     if (!NIL_P(accumulator))
-        G_RELATIVE(self, accumulator); /* FIXME */
+        rbgobj_add_relative(rb_signal, accumulator);
+    g_rclosure_attach(class_closure, rb_signal);
+    rbgobj_add_relative(self, rb_signal);
 
-    return rbgobj_signal_wrap(signal);
+    return rb_signal;
 }
 
 static void
@@ -258,7 +320,7 @@ gobj_s_signal(VALUE self, VALUE name)
     if (!sig_id)
         rb_raise(eNoSignalError, "no such signal: %s", sig_name);
 
-    return rbgobj_signal_wrap(sig_id);
+    return rbgobj_signal_new(sig_id);
 }
 
 static VALUE
@@ -309,9 +371,31 @@ gobj_sig_connect_impl(gboolean after, int argc, VALUE *argv, VALUE self)
     if (!g_signal_parse_name(sig_name, CLASS2GTYPE(CLASS_OF(self)), &signal_id, &detail, TRUE))
         rb_raise(eNoSignalError, "no such signal: %s", sig_name);
 
-    func = rb_block_proc();
-    rclosure = g_rclosure_new(func, rest, 
-                              rbgobj_get_signal_func(signal_id));
+    {
+        ID id_create_signal_handler;
+        const char *normalized_signal_name;
+
+        CONST_ID(id_create_signal_handler, "create_signal_handler");
+        normalized_signal_name = g_signal_name(signal_id);
+        func = rb_funcall(self,
+                          id_create_signal_handler,
+                          2,
+                          rb_str_new_cstr(normalized_signal_name),
+                          rb_block_proc());
+    }
+    {
+        RGClosureCallFunc call_func;
+        call_func = rbgobj_get_signal_call_func(signal_id);
+        if (call_func) {
+            rclosure = g_rclosure_new_call(func,
+                                           rest,
+                                           call_func);
+        } else {
+            rclosure = g_rclosure_new(func,
+                                      rest,
+                                      rbgobj_get_signal_func(signal_id));
+        }
+    }
     g_rclosure_attach((GClosure *)rclosure, self);
     g_object = RVAL2GOBJ(self);
     tag = g_strdup_printf("%s::%s",
@@ -344,7 +428,7 @@ gobj_sig_get_invocation_hint(VALUE self)
     GSignalInvocationHint* hint;
     hint = g_signal_get_invocation_hint(RVAL2GOBJ(self));
     return rb_ary_new3(3,
-                       rbgobj_signal_wrap(hint->signal_id),
+                       rbgobj_signal_new(hint->signal_id),
                        hint->detail ? rb_str_new2(g_quark_to_string(hint->detail)) : Qnil,
                        INT2NUM(hint->run_type));
 }
@@ -360,8 +444,9 @@ struct emit_arg {
 };
 
 static VALUE
-emit_body(struct emit_arg* arg)
+emit_body(VALUE rb_arg)
 {
+    struct emit_arg *arg = (struct emit_arg *)rb_arg;
     GValue param = G_VALUE_INIT;
 
     g_value_init(&param, G_TYPE_FROM_INSTANCE(RVAL2GOBJ(arg->self)));
@@ -405,8 +490,9 @@ emit_body(struct emit_arg* arg)
 }
 
 static VALUE
-emit_ensure(struct emit_arg* arg)
+emit_ensure(VALUE rb_arg)
 {
+    struct emit_arg *arg = (struct emit_arg *)rb_arg;
     g_value_array_free(arg->instance_and_params);
     return Qnil;
 }
@@ -539,8 +625,9 @@ guint    g_signal_handlers_disconnect_matched (gpointer         instance,
 #endif
 
 static VALUE
-chain_from_overridden_body(struct emit_arg* arg)
+chain_from_overridden_body(VALUE rb_arg)
 {
+    struct emit_arg *arg = (struct emit_arg *)rb_arg;
     g_value_init(arg->instance_and_params->values,
                  G_TYPE_FROM_INSTANCE(RVAL2GOBJ(arg->self)));
     rbgobj_rvalue_to_gvalue(arg->self, arg->instance_and_params->values);
@@ -645,71 +732,54 @@ gobj_s_method_added(VALUE klass, VALUE id)
 VALUE
 rbgobj_signal_wrap(guint sig_id)
 {
-    VALUE result;
-    GSignalQuery* query;
-
-    result = Data_Make_Struct(RG_TARGET_NAMESPACE, GSignalQuery, NULL, free, query);
-    g_signal_query(sig_id, query);
-    return result;
+    return rbgobj_signal_new(sig_id);
 }
 
 static VALUE
 rg_id(VALUE self)
 {
-    GSignalQuery* query;
-    Data_Get_Struct(self, GSignalQuery, query);
-    return UINT2NUM(query->signal_id);
+    return UINT2NUM(rbgobj_signal_get_raw(self)->signal_id);
 }
 
 static VALUE
 rg_name(VALUE self)
 {
-    GSignalQuery* query;
-    Data_Get_Struct(self, GSignalQuery, query);
-    return rb_str_new2(query->signal_name);
+    return CSTR2RVAL(rbgobj_signal_get_raw(self)->signal_name);
 }
 
 static VALUE
 rg_itype(VALUE self)
 {
-    GSignalQuery* query;
-    Data_Get_Struct(self, GSignalQuery, query);
-    return rbgobj_gtype_new(query->itype);
+    return rbgobj_gtype_new(rbgobj_signal_get_raw(self)->itype);
 }
 
 static VALUE
 rg_owner(VALUE self)
 {
-    GSignalQuery* query;
-    Data_Get_Struct(self, GSignalQuery, query);
-    return GTYPE2CLASS(query->itype);
+    return GTYPE2CLASS(rbgobj_signal_get_raw(self)->itype);
 }
 
 static VALUE
 rg_return_type(VALUE self)
 {
-    GSignalQuery* query;
-    Data_Get_Struct(self, GSignalQuery, query);
-    return rbgobj_gtype_new(query->return_type);
+    return rbgobj_gtype_new(rbgobj_signal_get_raw(self)->return_type);
 }
 
 static VALUE
 rg_flags(VALUE self)
 {
-    GSignalQuery* query;
-    Data_Get_Struct(self, GSignalQuery, query);
-    return UINT2NUM(query->signal_flags);
+    return UINT2NUM(rbgobj_signal_get_raw(self)->signal_flags);
 }
 
 static VALUE
 rg_param_types(VALUE self)
 {
-    GSignalQuery* query;
+    GSignalQuery *query;
     VALUE result;
     guint i;
-    Data_Get_Struct(self, GSignalQuery, query);
 
-    result = rb_ary_new2(query->n_params);
+    query = rbgobj_signal_get_raw(self);
+    result = rb_ary_new_capa(query->n_params);
     for (i = 0; i < query->n_params; i++)
         rb_ary_store(result, i, rbgobj_gtype_new(query->param_types[i]));
 
@@ -719,31 +789,19 @@ rg_param_types(VALUE self)
 static VALUE
 rg_inspect(VALUE self)
 {
-    GSignalQuery* query;
-    gchar* s;
-    VALUE result, v;
-
-    Data_Get_Struct(self, GSignalQuery, query);
-
-    v = rb_inspect(GTYPE2CLASS(query->itype));
-
-    s = g_strdup_printf("#<%s: %s#%s>",
-                        rb_class2name(CLASS_OF(self)),
-                        StringValuePtr(v),
-                        query->signal_name);
-    result = rb_str_new2(s);
-    g_free(s);
-
-    return result;
+    GSignalQuery *query;
+    query = rbgobj_signal_get_raw(self);
+    return rb_sprintf("#<%" PRIsVALUE " %+" PRIsVALUE "#%s>",
+                      CLASS_OF(self),
+                      GTYPE2CLASS(query->itype),
+                      query->signal_name);
 }
 
 #define query_is_flag(flag) \
     static VALUE \
     query_is_##flag(VALUE self) \
     { \
-        GSignalQuery* query; \
-        Data_Get_Struct(self, GSignalQuery, query); \
-        return CBOOL2RVAL(query->signal_flags & flag); \
+        return CBOOL2RVAL(rbgobj_signal_get_raw(self)->signal_flags & flag); \
     }
 
 query_is_flag(G_SIGNAL_RUN_FIRST)
@@ -760,8 +818,8 @@ hook_func(GSignalInvocationHint* ihint,
           const GValue*          param_values,
           gpointer               data)
 {
-    GClosure* closure = data;
-    GValue ret_val =G_VALUE_INIT;
+    GClosure *closure = data;
+    GValue ret_val = G_VALUE_INIT;
     gboolean ret;
 
     g_value_init(&ret_val, G_TYPE_BOOLEAN);
@@ -777,7 +835,7 @@ g_signal_add_emission_hook_closure (guint     signal_id,
                                     GQuark    detail,
                                     GClosure* closure)
 {
-    guint hook_id;
+    gulong hook_id;
     g_closure_ref(closure);
     g_closure_sink(closure);
     hook_id = g_signal_add_emission_hook(signal_id, detail,
@@ -789,13 +847,13 @@ g_signal_add_emission_hook_closure (guint     signal_id,
 static VALUE
 rg_add_emission_hook(int argc, VALUE* argv, VALUE self)
 {
-    GSignalQuery* query;
+    GSignalQuery *query;
     VALUE proc;
-    guint hook_id;
+    gulong hook_id;
     GQuark detail = 0;
     GClosure* closure;
 
-    Data_Get_Struct(self, GSignalQuery, query);
+    query = rbgobj_signal_get_raw(self);
 
     if (query->signal_flags & G_SIGNAL_DETAILED) {
         VALUE detail_obj;
@@ -803,7 +861,7 @@ rg_add_emission_hook(int argc, VALUE* argv, VALUE self)
             if (SYMBOL_P(detail_obj))
                 detail = g_quark_from_string(rb_id2name(SYM2ID(detail_obj)));
             else
-                detail = g_quark_from_string(StringValuePtr(detail_obj));
+                detail = g_quark_from_string(RVAL2CSTR(detail_obj));
         }
     } else {
         rb_scan_args(argc, argv, "00&", &proc);
@@ -819,9 +877,8 @@ rg_add_emission_hook(int argc, VALUE* argv, VALUE self)
 static VALUE
 rg_remove_emission_hook(VALUE self, VALUE hook_id)
 {
-    GSignalQuery* query;
-    Data_Get_Struct(self, GSignalQuery, query);
-    g_signal_remove_emission_hook(query->signal_id, NUM2ULONG(hook_id));
+    g_signal_remove_emission_hook(rbgobj_signal_get_raw(self)->signal_id,
+                                  NUM2ULONG(hook_id));
     return Qnil;
 }
 
@@ -936,7 +993,11 @@ Init_gobject_gsignal(void)
     signal_func_table = rb_hash_new();
     rb_global_variable(&signal_func_table);
 
-    rbg_define_method(mMetaInterface, "signal_new", gobj_s_signal_new, -1);
+    signal_call_func_table = rb_hash_new();
+    rb_global_variable(&signal_call_func_table);
+
+    rbg_define_method(mMetaInterface, "define_signal", gobj_s_define_signal, -1);
+    rb_define_alias(mMetaInterface, "signal_new", "define_signal");
     rbg_define_method(mMetaInterface, "signals", gobj_s_signals, -1);
     rbg_define_method(mMetaInterface, "signal", gobj_s_signal, 1);
 
